import logging
from typing import Any, Callable, Dict, List, Set, Union
from urllib.parse import urlencode

from django.apps import apps
from django.contrib.admin import ListFilter
from django.contrib.admin.helpers import AdminForm
from django.contrib.auth.models import AbstractUser
from django.db.models.base import Model, ModelBase
from django.db.models.options import Options
from django.urls import NoReverseMatch, reverse
from django.utils.translation import gettext

logger = logging.getLogger(__name__)


def order_with_respect_to(original: List, reference: List, getter: Callable = lambda x: x) -> List:
    """
    Order a list based on the location of items in the reference list, optionally, use a getter to pull values out of
    the first list
    """
    ranking = []
    max_num = len(original)
    for item in original:
        try:
            pos = reference.index(getter(item))
        except ValueError:
            pos = max_num

        ranking.append(pos)

    return [y for x, y in sorted(zip(ranking, original), key=lambda x: x[0])]


def get_admin_url(instance: Any, admin_site: str = 'admin', from_app: bool = False, **kwargs: str) -> str:
    """
    Return the admin URL for the given instance, model class or <app>.<model> string
    """
    url = '#'

    try:

        if type(instance) is str:
            app_label, model_name = instance.split('.')
            model_name = model_name.lower()
            url = reverse(
                'admin:{app_label}_{model_name}_changelist'.format(app_label=app_label, model_name=model_name),
                current_app=admin_site,
            )

        # Model class
        elif instance.__class__ == ModelBase:
            app_label, model_name = instance._meta.app_label, instance._meta.model_name
            url = reverse(
                'admin:{app_label}_{model_name}_changelist'.format(app_label=app_label, model_name=model_name),
                current_app=admin_site,
            )

        # Model instance
        elif instance.__class__.__class__ == ModelBase and isinstance(instance, instance.__class__):
            app_label, model_name = instance._meta.app_label, instance._meta.model_name
            url = reverse(
                'admin:{app_label}_{model_name}_change'.format(app_label=app_label, model_name=model_name),
                args=(instance.pk,),
                current_app=admin_site,
            )

    except (NoReverseMatch, ValueError):
        # If we are not walking through the models within an app, let the user know this url cant be reversed
        if not from_app:
            logger.warning(gettext('Could not reverse url from {instance}'.format(instance=instance)))

    if kwargs:
        url += '?{params}'.format(params=urlencode(kwargs))

    return url


def get_filter_id(spec: ListFilter) -> str:
    return getattr(spec, 'field_path', getattr(spec, 'parameter_name', spec.title))


def get_custom_url(url: str, admin_site: str = 'admin') -> str:
    """
    Take in a custom url, and try to reverse it
    """
    if not url:
        logger.warning('No url supplied in custom link')
        return '#'

    if '/' in url:
        return url
    try:
        url = reverse(url.lower(), current_app=admin_site)
    except NoReverseMatch:
        logger.warning('Couldnt reverse {url}'.format(url=url))
        url = '#' + url

    return url


def get_model_meta(model_str: str) -> Union[None, Options]:
    """
    Get model meta class
    """
    try:
        app, model = model_str.split('.')
        model_klass: Model = apps.get_registered_model(app, model)
        return model_klass._meta
    except (ValueError, LookupError):
        return None


def get_app_admin_urls(app: str, admin_site: str = 'admin') -> List[Dict]:
    """
    For the given app string, get links to all the app models admin views
    """
    if app not in apps.app_configs:
        logger.warning('{app} not found when generating links'.format(app=app))
        return []

    models = []
    for model in apps.app_configs[app].get_models():
        url = get_admin_url(model, admin_site=admin_site, from_app=True)

        # We have no admin class
        if url == '#':
            continue

        models.append({
            'url': url,
            'model': '{app}.{model}'.format(app=model._meta.app_label, model=model._meta.model_name),
            'name': model._meta.verbose_name_plural.title(),
        })

    return models


def get_view_permissions(user: AbstractUser) -> Set[str]:
    """
    Get model names based on a users view/change permissions
    """
    perms = user.get_all_permissions()
    # the perm codenames should always be lower case
    lower_perms = []
    for perm in perms:
        app, perm_codename = perm.split('.')
        lower_perms.append('{app}.{perm_codename}'.format(app=app, perm_codename=perm_codename.lower()))
    return {x.replace('view_', '') for x in lower_perms if 'view' in x or 'change' in x}


def make_menu(
    user: AbstractUser,
    links: List[Dict],
    options: Dict,
    allow_appmenus: bool = True,
    admin_site: str = 'admin'
) -> List[Dict]:
    """
    Make a menu from a list of user supplied links
    """
    if not user:
        return []

    model_permissions = get_view_permissions(user)

    menu = []
    for link in links:

        perm_matches = []
        for perm in link.get('permissions', []):
            perm_matches.append(user.has_perm(perm))

        if not all(perm_matches):
            continue

        # Url links
        if 'url' in link:
            menu.append({
                'name': link.get('name', 'unspecified'),
                'url': get_custom_url(link['url'], admin_site=admin_site),
                'children': None,
                'new_window': link.get('new_window', False),
                'icon': link.get('icon', options['default_icon_children']),
            })

        # Model links
        elif 'model' in link:
            if link['model'].lower() not in model_permissions:
                continue

            _meta = get_model_meta(link['model'])

            name = _meta.verbose_name_plural.title() if _meta else link['model']
            menu.append({
                'name': name,
                'url': get_admin_url(link['model'], admin_site=admin_site),
                'children': [],
                'new_window': link.get('new_window', False),
                'icon': options['icons'].get(link['model'], options['default_icon_children']),
            })

        # App links
        elif 'app' in link and allow_appmenus:
            app_admin_urls = get_app_admin_urls(link['app'], admin_site=admin_site)
            children = [{
                'name': child.get('verbose_name', child['name']),
                'url': child['url'],
                'children': None
            } for child in app_admin_urls if child['model'] in model_permissions]
            if len(children) == 0:
                continue

            menu.append({
                'name': getattr(apps.app_configs[link['app']], 'verbose_name', link['app']).title(),
                'url': '#',
                'children': children,
                'icon': options['icons'].get(link['app'], options['default_icon_children']),
            })

    return menu


def has_fieldsets_check(adminform: AdminForm) -> bool:
    fieldsets = adminform.fieldsets
    if not fieldsets or (len(fieldsets) == 1 and fieldsets[0][0] is None):
        return False
    return True


def attr(**kwargs) -> Callable:

    def decorator(func: Callable):
        for key, value in kwargs.items():
            setattr(func, key, value)
        return func

    return decorator



# Generated by Django 3.2.12 on 2022-03-27 07:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("books", "0005_book_last_print"),
    ]

    operations = [
        migrations.AlterField(
            model_name="author",
            name="id",
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
        ),
        migrations.AlterField(
            model_name="book",
            name="id",
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
        ),
        migrations.AlterField(
            model_name="genre",
            name="id",
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
        ),
    ]



from django.contrib import admin
from django.contrib.admin.models import LogEntry
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.models import User
from django.utils.html import format_html
from django.utils.timesince import timesince

from richmin.utils import attr
from .models import Author, Book, Genre
from ..loans.admin import BookLoanInline

admin.site.unregister(User)


class BooksInline(admin.TabularInline):
    model = Book


@admin.register(Book)
class BookAdmin(admin.ModelAdmin):
    fieldsets = (
        ('general', {
            'fields': ('title', 'author', 'library'),
            'description': 'General book fields'
        }),
        ('other', {
            'fields': ('genre', 'summary', 'isbn', 'published_on', 'pages')
        }),
    )
    raw_id_fields = ('author',)
    list_display = ('__str__', 'title', 'author', 'pages')
    readonly_fields = ('__str__',)
    list_per_page = 20
    list_max_show_all = 100
    list_editable = ('title',)
    search_fields = ('title', 'author__last_name')
    autocomplete_fields = ('genre',)
    date_hierarchy = 'published_on'
    save_as = True
    save_on_top = True
    inlines = (BookLoanInline,)

    actions_on_bottom = True

    # Order the sections within the change form
    richmin_section_order = ('book loans', 'general', 'other')


@admin.register(Author)
class AuthorAdmin(admin.ModelAdmin):
    list_display = ('last_name', 'first_name', 'date_of_birth', 'date_of_death')
    fields = ('first_name', 'last_name', ('date_of_birth', 'date_of_death'))
    inlines = (BooksInline,)


@admin.register(LogEntry)
class LogEntryAdmin(admin.ModelAdmin):
    list_display = ('user', 'object', 'action_flag', 'change_message', 'modified')
    readonly_fields = ['object', 'modified']
    search_fields = ('user__email',)
    date_hierarchy = 'action_time'
    list_filter = ('action_flag', 'content_type__model')
    list_per_page = 20

    def object(self, obj):
        url = obj.get_admin_url()
        return format_html(
            '<a href="{url}">{obj} [{model}]</a>'.format(url=url, obj=obj.object_repr, model=obj.content_type.model)
        )

    @attr(admin_order_field='action_time')
    def modified(self, obj):
        if not obj.action_time:
            return 'Never'
        return '{} ago'.format(timesince(obj.action_time))


@admin.register(User)
class CustomUserAdmin(UserAdmin):

    def get_queryset(self, request):
        """
        Remove our test user from the admin, so it cant be messed with
        """
        return super().get_queryset(request).exclude(username='test@test.com')


@admin.register(Genre)
class GenreAdmin(admin.ModelAdmin):
    search_fields = ('name',)



from uuid import uuid4

from django.conf import settings
from django.db import models
from django.utils import timezone


class Library(models.Model):
    name = models.CharField(max_length=255)
    address = models.CharField(max_length=255)
    librarian = models.OneToOneField(settings.AUTH_USER_MODEL, related_name='library', on_delete=models.CASCADE)

    class Meta:
        verbose_name_plural = 'Libraries'

    def __str__(self):
        return self.address


class BookLoan(models.Model):
    id = models.UUIDField(
        primary_key=True,
        default=uuid4,
        help_text='Unique ID for this particular book across whole library',
    )
    book = models.ForeignKey('books.Book', on_delete=models.SET_NULL, null=True)
    imprint = models.CharField(max_length=200)
    loan_start = models.DateTimeField()
    due_back = models.DateField(null=True, blank=True)

    duration = models.DurationField(blank=True)

    LOAN_STATUS = (
        ('m', 'Maintenance'),
        ('o', 'On loan'),
        ('a', 'Available'),
        ('r', 'Reserved'),
    )
    borrower = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True, blank=True)
    status = models.CharField(
        max_length=1,
        choices=LOAN_STATUS,
        blank=True,
        default='m',
        help_text='Book availability',
    )

    class Meta:
        ordering = ('due_back',)

    def __str__(self):
        return '{}, ({})'.format(self.id, self.book.title)

    def save(self, **kwargs):
        self.duration = self.loan_start - timezone.now()
        super().save(**kwargs)



import copy

import pytest
from django.urls import reverse

from .test_app.library.factories import BookFactory


@pytest.fixture
def change_form_context(admin_client):
    book = BookFactory()
    url = reverse(
        'admin:books_book_change', args=(book.pk,)
    )
    response = admin_client.get(url)
    yield response.context


@pytest.fixture(scope='function')
def custom_richmin_settings(settings):
    original_settings = copy.deepcopy(settings.RICHMIN_SETTINGS)
    settings.RICHMIN_SETTINGS = original_settings
    yield original_settings



import os
from pathlib import Path
from typing import Any, Dict

from django.conf.global_settings import LANGUAGES as DJANGO_LANGUAGES

###################
# Django Settings #
###################
BASE_DIR = Path(__file__).resolve().parent
SECRET_KEY = os.getenv('DJANGO_SECRET_KEY', 'not-secret-at-all')
DEBUG = bool(int(os.getenv('DEBUG', 1)))
TEST = os.getenv('FAIL_INVALID_TEMPLATE_VARS')

PREFIX = '' if os.getenv('STANDALONE') else ''

ALLOWED_HOSTS = ['*']
INSTALLED_APPS = [
    # Keep this above 'django.contrib.admin'
    'richmin',
    # Django apps
    'django.contrib.admin',
    'django.contrib.admindocs',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    # Our apps
    '{}library.books.apps.BooksConfig'.format(PREFIX),
    '{}library.loans.apps.LoansConfig'.format(PREFIX),
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.admindocs.middleware.XViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = '{}library.urls'.format(PREFIX)

WSGI_APPLICATION = '{}library.wsgi.application'.format(PREFIX)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler'
        }
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'INFO'
        }
    },
}

TEMPLATES = [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'DIRS': [],
    'APP_DIRS': True,
    'OPTIONS': {
        'context_processors': [
            'django.template.context_processors.debug',
            'django.template.context_processors.request',
            'django.contrib.auth.context_processors.auth',
            'django.contrib.messages.context_processors.messages',
        ],
    },
}]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    }
}

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator'
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator'
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator'
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator'
    },
]

LANGUAGE_CODE = 'en'
TIME_ZONE = 'Europe/London'
USE_I18N = True
LOCALE_PATHS = (os.path.join(BASE_DIR, 'locale'),)
USE_L10N = True
USE_TZ = True

# English default
LANGUAGES = DJANGO_LANGUAGES

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'static')

MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'

if DEBUG and not TEST:
    os.environ.setdefault('WERKZEUG_DEBUG_PIN', 'off')
    INSTALLED_APPS.extend(['debug_toolbar', 'django_extensions'])
    MIDDLEWARE.insert(0, 'debug_toolbar.middleware.DebugToolbarMiddleware')
    DEBUG_TOOLBAR_CONFIG = {'SHOW_TOOLBAR_CALLBACK': lambda _: False}

if not DEBUG and not TEST:
    MIDDLEWARE.insert(1, 'whitenoise.middleware.WhiteNoiseMiddleware')
    STATICFILES_STORAGE = 'whitenoise.storage.CompressedManifestStaticFilesStorage'

########################
# Third party settings #
########################
RICHMIN_SETTINGS: Dict[str, Any] = {
    # title of the window (Will default to current_admin_site.site_title if absent or None)
    'site_title': 'Library Admin',
    # Title on the brand, and login screen (19 chars max) (defaults to current_admin_site.site_header if absent or None)
    'site_header': 'Library',
    # Logo to use for your site, must be present in static files, used for brand on top left
    'site_logo': 'books/img/logo.png',
    # Relative path to logo for your site, used for login logo (must be present in static files. Defaults to site_logo)
    'login_logo': 'books/img/logo-login.png',
    # Logo to use for login form in dark themes (must be present in static files. Defaults to login_logo)
    'login_logo_dark': 'books/img/logo-login-dark-mode.png',
    # CSS classes that are applied to the logo
    'site_logo_classes': None,
    # Relative path to a favicon for your site, will default to site_logo if absent (ideally 32x32 px)
    'site_icon': 'books/img/icon.png',
    # Welcome text on the login screen
    'welcome_sign': 'Welcome to the library',
    # Copyright on the footer
    'copyright': 'Richmin Library',
    # List of model admins to search from the search bar, search bar omitted if excluded
    # If you want to use a single search field you dont need to use a list, you can use a simple string
    'search_model': ['auth.User', 'auth.Group'],
    # Field name on user model that contains avatar ImageField/URLField/Charfield or a callable that receives the user
    'user_avatar': None,
    ############
    # Top Menu #
    ############
    # Links to put along the top menu
    'topmenu_links': [
        # Url that gets reversed (Permissions can be added)
        {
            'name': 'Home',
            'url': 'admin:index',
            'permissions': ['auth.view_user']
        },
        # external url that opens in a new window (Permissions can be added)
        {
            'name': 'Support',
            'url': 'https://github.com/mymy47/django-richmin',
            'new_window': True
        },
        # model admin to link to (Permissions checked against model)
        {
            'model': 'auth.User'
        },
        # App with dropdown menu to all its models pages (Permissions checked against models)
        {
            'app': 'books'
        },
        {
            'app': 'loans'
        },
    ],
    #############
    # User Menu #
    #############
    # Additional links to include in the user menu on the top right ('app' url type is not allowed)
    'usermenu_links': [
        {
            'name': 'Support',
            'url': 'https://github.com/mymy47/django-richmin',
            'new_window': True
        },
        {
            'model': 'auth.user'
        },
    ],
    #############
    # Side Menu #
    #############
    # Whether to display the side menu
    'show_sidebar': True,
    # Whether to aut expand the menu
    'navigation_expanded': True,
    # Hide these apps when generating side menu e.g (auth)
    'hide_apps': [],
    # Hide these models when generating side menu (e.g auth.user)
    'hide_models': [],
    # List of apps to base side menu (app or model) ordering off of
    'order_with_respect_to': ['Make Messages', 'auth', 'books', 'books.author', 'books.book', 'loans'],
    # Custom links to append to app groups, keyed on app name
    'custom_links': {
        'loans': [
            {
                'name': 'Make Messages',
                'url': 'make_messages',
                'icon': 'fas fa-comments',
                'permissions': ['loans.view_loan'],
            },
            {
                'name': 'Custom View',
                'url': 'admin:custom_view',
                'icon': 'fas fa-box-open'
            },
        ]
    },
    # Custom icons for side menu apps/models See the link below
    # https://fontawesome.com/icons?d=gallery&m=free&v=5.0.0,5.0.1,5.0.10,5.0.11,5.0.12,5.0.13,5.0.2,5.0.3,5.0.4,5.0.5,5.0.6,5.0.7,5.0.8,5.0.9,5.1.0,
    # 5.1.1,5.2.0,5.3.0,5.3.1,5.4.0,5.4.1,5.4.2,5.13.0,5.12.0,5.11.2,5.11.1,5.10.0,5.9.0,5.8.2,5.8.1,5.7.2,5.7.1,5.7.0,5.6.3,5.5.0,5.4.2
    # for the full list of 5.13.0 free icon classes
    'icons': {
        'auth': 'fas fa-users-cog',
        'auth.user': 'fas fa-user',
        'auth.Group': 'fas fa-users',
        'admin.LogEntry': 'fas fa-file',
        'books.Author': 'fas fa-user',
        'books.Book': 'fas fa-book',
        'books.Genre': 'fas fa-photo-video',
        'loans.BookLoan': 'fas fa-book-open',
        'loans.Library': 'fas fa-book-reader',
    },
    # Icons that are used when one is not manually specified
    'default_icon_parents': 'fas fa-chevron-circle-right',
    'default_icon_children': 'fas fa-circle',
    #################
    # Related Modal #
    #################
    # Activate Bootstrap modal
    'related_modal_active': False,
    #############
    # UI Tweaks #
    #############
    # Relative paths to custom CSS/JS scripts (must be present in static files)
    'custom_css': None,
    'custom_js': None,
    # Whether to show the UI customizer on the sidebar
    'show_ui_builder': True,
    ###############
    # Change view #
    ###############
    # Render out the change view as a single form, or in tabs, current options are
    # - single
    # - horizontal_tabs (default)
    # - vertical_tabs
    # - collapsible
    # - carousel
    'changeform_format': 'horizontal_tabs',
    # override change forms on a per modeladmin basis
    'changeform_format_overrides': {
        'auth.user': 'collapsible',
        'auth.group': 'vertical_tabs'
    },
    # Add a language dropdown into the admin
    'language_chooser': True,
}

if not DEBUG and not TEST:
    RICHMIN_SETTINGS['welcome_sign'] = 'Username: test@test.com, Password: test (Data resets nightly)'

RICHMIN_UI_TWEAKS = {
    'navbar_small_text': False,
    'footer_small_text': False,
    'body_small_text': False,
    'brand_small_text': False,
    'brand_colour': False,
    'accent': 'accent-primary',
    'navbar': 'navbar-white navbar-light',
    'no_navbar_border': False,
    'navbar_fixed': False,
    'layout_boxed': False,
    'footer_fixed': False,
    'sidebar_fixed': False,
    'sidebar': 'sidebar-dark-primary',
    'sidebar_nav_small_text': False,
    'sidebar_disable_expand': False,
    'sidebar_nav_child_indent': False,
    'sidebar_nav_compact_style': False,
    'sidebar_nav_legacy_style': False,
    'sidebar_nav_flat_style': False,
    'theme': 'default',
    'dark_mode_theme': 'darkly',
    'button_classes': {
        'primary': 'btn-outline-primary',
        'secondary': 'btn-outline-secondary',
        'info': 'btn-outline-info',
        'warning': 'btn-outline-warning',
        'danger': 'btn-outline-danger',
        'success': 'btn-outline-success',
    },
}

################
# App settings #
################



import copy
import logging
from typing import Any, Dict

from django.conf import settings
from django.templatetags.static import static

from .utils import get_admin_url, get_model_meta

logger = logging.getLogger(__name__)

DEFAULT_SETTINGS: Dict[str, Any] = {
    # title of the window (Will default to current_admin_site.site_title)
    'site_title': None,
    # Title on the login screen (19 chars max) (will default to current_admin_site.site_header)
    'site_header': None,
    # Title on the brand (19 chars max) (will default to current_admin_site.site_header)
    'site_brand': None,
    # Relative path to logo for your site, used for brand on top left (must be present in static files)
    'site_logo': 'vendor/adminlte/img/AdminLTELogo.png',
    # Relative path to logo for your site, used for login logo (must be present in static files. Defaults to site_logo)
    'login_logo': None,
    # Logo to use for login form in dark themes (must be present in static files. Defaults to login_logo)
    'login_logo_dark': None,
    # CSS classes that are applied to the logo
    'site_logo_classes': 'img-circle',
    # Relative path to a favicon for your site, will default to site_logo if absent (ideally 32x32 px)
    'site_icon': None,
    # Welcome text on the login screen
    'welcome_sign': 'Welcome',
    # Copyright on the footer
    'copyright': '',
    # The model admin to search from the search bar, search bar omitted if excluded
    'search_model': None,
    # Field name on user model that contains avatar ImageField/URLField/Charfield or a callable that receives the user
    'user_avatar': None,
    ############
    # Top Menu #
    ############
    # Links to put along the nav bar
    'topmenu_links': [],
    #############
    # User Menu #
    #############
    # Additional links to include in the user menu on the top right ('app' url type is not allowed)
    'usermenu_links': [],
    #############
    # Side Menu #
    #############
    # Whether to display the side menu
    'show_sidebar': True,
    # Whether to aut expand the menu
    'navigation_expanded': True,
    # Hide these apps when generating side menu e.g (auth)
    'hide_apps': [],
    # Hide these models when generating side menu (e.g auth.user)
    'hide_models': [],
    # List of apps to base side menu ordering off of
    'order_with_respect_to': [],
    # Custom links to append to side menu app groups, keyed on app name
    'custom_links': {},
    # Custom icons for side menu apps/models See the link below
    # https://fontawesome.com/icons?d=gallery&m=free&v=5.0.0,5.0.1,5.0.10,5.0.11,5.0.12,5.0.13,5.0.2,5.0.3,5.0.4,5.0.5,5.0.6,5.0.7,5.0.8,5.0.9,5.1.0,
    # 5.1.1,5.2.0,5.3.0,5.3.1,5.4.0,5.4.1,5.4.2,5.13.0,5.12.0,5.11.2,5.11.1,5.10.0,5.9.0,5.8.2,5.8.1,5.7.2,5.7.1,5.7.0,5.6.3,5.5.0,5.4.2
    # for the full list of 5.13.0 free icon classes
    'icons': {
        'auth': 'fas fa-users-cog',
        'auth.user': 'fas fa-user',
        'auth.Group': 'fas fa-users'
    },
    # Icons that are used when one is not manually specified
    'default_icon_parents': 'fas fa-chevron-circle-right',
    'default_icon_children': 'fas fa-circle',
    #################
    # Related Modal #
    #################
    # Activate Bootstrap modal
    'related_modal_active': False,
    #############
    # UI Tweaks #
    #############
    # Relative paths to custom CSS/JS scripts (must be present in static files)
    'custom_css': None,
    'custom_js': None,
    # Whether to link font from fonts.googleapis.com (use custom_css to supply font otherwise)
    'use_google_fonts_cdn': True,
    # Whether to show the UI customizer on the sidebar
    'show_ui_builder': False,
    ###############
    # Change view #
    ###############
    # Render out the change view as a single form, or in tabs, current options are
    # - single
    # - horizontal_tabs (default)
    # - vertical_tabs
    # - collapsible
    # - carousel
    'changeform_format': 'horizontal_tabs',
    # override change forms on a per modeladmin basis
    'changeform_format_overrides': {},
    # Add a language dropdown into the admin
    'language_chooser': False,
}

#######################################
# Currently available UI tweaks       #
# Use the UI builder to generate this #
#######################################

DEFAULT_UI_TWEAKS: Dict[str, Any] = {
    # Small text on the top navbar
    'navbar_small_text': False,
    # Small text on the footer
    'footer_small_text': False,
    # Small text everywhere
    'body_small_text': True,
    # Small text on the brand/logo
    'brand_small_text': False,
    # brand/logo background colour
    'brand_colour': 'navbar-gray',
    # Link colour
    'accent': 'accent-primary',
    # topmenu colour
    'navbar': 'navbar-white navbar-light',
    # topmenu border
    'no_navbar_border': False,
    # Make the top navbar sticky, keeping it in view as you scroll
    'navbar_fixed': True,
    # Whether to constrain the page to a box (leaving big margins at the side)
    'layout_boxed': False,
    # Make the footer sticky, keeping it in view all the time
    'footer_fixed': False,
    # Make the sidebar sticky, keeping it in view as you scroll
    'sidebar_fixed': True,
    # sidemenu colour
    'sidebar': 'sidebar-dark-info',
    # sidemenu small text
    'sidebar_nav_small_text': False,
    # Disable expanding on hover of collapsed sidebar
    'sidebar_disable_expand': False,
    # Indent child menu items on sidebar
    'sidebar_nav_child_indent': False,
    # Use a compact sidebar
    'sidebar_nav_compact_style': False,
    # Use the AdminLTE2 style sidebar
    'sidebar_nav_legacy_style': False,
    # Use a flat style sidebar
    'sidebar_nav_flat_style': False,
    # Bootstrap theme to use (default, or from bootswatch, see THEMES below)
    'theme': 'default',
    # Theme to use instead if the user has opted for dark mode (e.g darkly/cyborg/slate/solar/superhero)
    'dark_mode_theme': None,
    # The classes/styles to use with buttons
    'button_classes': {
        'primary': 'btn-outline-primary',
        'secondary': 'btn-outline-secondary',
        'info': 'btn-info',
        'warning': 'btn-warning',
        'danger': 'btn-danger',
        'success': 'btn-success'
    },
    'actions_sticky_top': True
}

THEMES = {
    # light themes
    'default': 'vendor/bootswatch/default/bootstrap.min.css',
    'cerulean': 'vendor/bootswatch/cerulean/bootstrap.min.css',
    'cosmo': 'vendor/bootswatch/cosmo/bootstrap.min.css',
    'flatly': 'vendor/bootswatch/flatly/bootstrap.min.css',
    'journal': 'vendor/bootswatch/journal/bootstrap.min.css',
    'litera': 'vendor/bootswatch/litera/bootstrap.min.css',
    'lumen': 'vendor/bootswatch/lumen/bootstrap.min.css',
    'lux': 'vendor/bootswatch/lux/bootstrap.min.css',
    'materia': 'vendor/bootswatch/materia/bootstrap.min.css',
    'minty': 'vendor/bootswatch/minty/bootstrap.min.css',
    'pulse': 'vendor/bootswatch/pulse/bootstrap.min.css',
    'sandstone': 'vendor/bootswatch/sandstone/bootstrap.min.css',
    'simplex': 'vendor/bootswatch/simplex/bootstrap.min.css',
    'sketchy': 'vendor/bootswatch/sketchy/bootstrap.min.css',
    'spacelab': 'vendor/bootswatch/spacelab/bootstrap.min.css',
    'united': 'vendor/bootswatch/united/bootstrap.min.css',
    'yeti': 'vendor/bootswatch/yeti/bootstrap.min.css',
    # dark themes
    'darkly': 'vendor/bootswatch/darkly/bootstrap.min.css',
    'cyborg': 'vendor/bootswatch/cyborg/bootstrap.min.css',
    'slate': 'vendor/bootswatch/slate/bootstrap.min.css',
    'solar': 'vendor/bootswatch/solar/bootstrap.min.css',
    'superhero': 'vendor/bootswatch/superhero/bootstrap.min.css',
}

DARK_THEMES = ('darkly', 'cyborg', 'slate', 'solar', 'superhero')

CHANGEFORM_TEMPLATES = {
    'single': 'richmin/includes/single.html',
    'carousel': 'richmin/includes/carousel.html',
    'collapsible': 'richmin/includes/collapsible.html',
    'horizontal_tabs': 'richmin/includes/horizontal_tabs.html',
    'vertical_tabs': 'richmin/includes/vertical_tabs.html',
}


def get_search_model_string(search_model: str) -> str:
    """
    Get a search model string for reversing an admin url.

    Ensure the model name is lower cased but remain the app name untouched.
    """

    app, model_name = search_model.split('.')
    return '{app}.{model_name}'.format(app=app, model_name=model_name.lower())


def get_settings() -> Dict:
    richmin_settings = copy.deepcopy(DEFAULT_SETTINGS)
    user_settings = {x: y for x, y in getattr(settings, 'RICHMIN_SETTINGS', {}).items() if y is not None}
    richmin_settings.update(user_settings)

    # Extract search model configuration from search_model setting
    if richmin_settings['search_model']:
        if not isinstance(richmin_settings['search_model'], list):
            richmin_settings['search_model'] = [richmin_settings['search_model']]

        richmin_settings['search_models_parsed'] = []
        for search_model in richmin_settings['search_model']:
            richmin_search_model = {}
            richmin_search_model['search_url'] = get_admin_url(get_search_model_string(search_model))

            model_meta = get_model_meta(search_model)
            if model_meta:
                richmin_search_model['search_name'] = model_meta.verbose_name_plural.title()
            else:
                richmin_search_model['search_name'] = search_model.split('.')[-1] + 's'
            richmin_settings['search_models_parsed'].append(richmin_search_model)

    # Deal with single strings in hide_apps/hide_models and make sure we lower case 'em
    if type(richmin_settings['hide_apps']) is str:
        richmin_settings['hide_apps'] = [richmin_settings['hide_apps']]
    richmin_settings['hide_apps'] = [x.lower() for x in richmin_settings['hide_apps']]

    if type(richmin_settings['hide_models']) is str:
        richmin_settings['hide_models'] = [richmin_settings['hide_models']]
    richmin_settings['hide_models'] = [x.lower() for x in richmin_settings['hide_models']]

    # Ensure icon model names and classes are lower case
    richmin_settings['icons'] = {x.lower(): y.lower() for x, y in richmin_settings.get('icons', {}).items()}

    # Default the site icon using the site logo
    richmin_settings['site_icon'] = richmin_settings['site_icon'] or richmin_settings['site_logo']

    # Default the login logo using the site logo
    richmin_settings['login_logo'] = richmin_settings['login_logo'] or richmin_settings['site_logo']

    # Default the login logo dark using the login logo
    richmin_settings['login_logo_dark'] = richmin_settings['login_logo_dark'] or richmin_settings['login_logo']

    # ensure all model names are lower cased
    richmin_settings['changeform_format_overrides'] = {
        x.lower(): y.lower()
        for x, y in richmin_settings.get('changeform_format_overrides', {}).items()
    }

    return richmin_settings


def get_ui_tweaks() -> Dict:
    raw_tweaks = copy.deepcopy(DEFAULT_UI_TWEAKS)
    raw_tweaks.update(getattr(settings, 'RICHMIN_UI_TWEAKS', {}))
    tweaks = {x: y for x, y in raw_tweaks.items() if y not in (None, '', False)}

    # These options don't work well together
    if tweaks.get('layout_boxed'):
        tweaks.pop('navbar_fixed', None)
        tweaks.pop('footer_fixed', None)

    bool_map = {
        'navbar_small_text': 'text-sm',
        'footer_small_text': 'text-sm',
        'body_small_text': 'text-sm',
        'brand_small_text': 'text-sm',
        'sidebar_nav_small_text': 'text-sm',
        'no_navbar_border': 'border-bottom-0',
        'sidebar_disable_expand': 'sidebar-no-expand',
        'sidebar_nav_child_indent': 'nav-child-indent',
        'sidebar_nav_compact_style': 'nav-compact',
        'sidebar_nav_legacy_style': 'nav-legacy',
        'sidebar_nav_flat_style': 'nav-flat',
        'layout_boxed': 'layout-boxed',
        'sidebar_fixed': 'layout-fixed',
        'navbar_fixed': 'layout-navbar-fixed',
        'footer_fixed': 'layout-footer-fixed',
        'actions_sticky_top': 'sticky-top',
    }

    for key, value in bool_map.items():
        if key in tweaks:
            tweaks[key] = value

    def classes(*args: str) -> str:
        return ' '.join([tweaks.get(arg, '') for arg in args]).strip()

    theme = tweaks['theme']
    if theme not in THEMES:
        logger.warning('{} not found in {}, using default'.format(theme, THEMES.keys()))
        theme = 'default'

    dark_mode_theme = tweaks.get('dark_mode_theme', None)
    if dark_mode_theme and dark_mode_theme not in DARK_THEMES:
        logger.warning('{} is not a dark theme, using darkly'.format(dark_mode_theme))
        dark_mode_theme = 'darkly'

    theme_body_classes = ' theme-{}'.format(theme)
    if theme in DARK_THEMES:
        theme_body_classes += ' dark-mode'

    ret = {
        'raw': raw_tweaks,
        'theme': {
            'name': theme,
            'src': static(THEMES[theme])
        },
        'sidebar_classes': classes('sidebar', 'sidebar_disable_expand'),
        'navbar_classes': classes('navbar', 'no_navbar_border', 'navbar_small_text'),
        'body_classes': classes(
            'accent', 'body_small_text', 'navbar_fixed', 'footer_fixed', 'sidebar_fixed', 'layout_boxed'
        ) + theme_body_classes,
        'actions_classes': classes('actions_sticky_top'),
        'sidebar_list_classes': classes(
            'sidebar_nav_small_text',
            'sidebar_nav_flat_style',
            'sidebar_nav_legacy_style',
            'sidebar_nav_child_indent',
            'sidebar_nav_compact_style',
        ),
        'brand_classes': classes('brand_small_text', 'brand_colour'),
        'footer_classes': classes('footer_small_text'),
        'button_classes': tweaks['button_classes'],
    }

    if dark_mode_theme:
        ret['dark_mode_theme'] = {'name': dark_mode_theme, 'src': static(THEMES[dark_mode_theme])}

    return ret



from django.contrib.auth.models import User
from django.db.models.signals import pre_save
from django.dispatch import receiver


@receiver(pre_save, sender=User)
def add_user_levels_actions(sender, instance, **kwargs):
    """
    Dont allow our test user to change their password
    """
    if instance.username == 'test@test.com':
        instance.set_password('test')



from django import forms
from django.forms.widgets import Select, SelectMultiple


class RichminSelect(Select):
    template_name = 'richmin/widgets/select.html'

    @property
    def media(self):
        return forms.Media(
            css={'all': ('vendor/select2/css/select2.min.css',)},
            js=('vendor/select2/js/select2.min.js',),
        )


class RichminSelectMultiple(SelectMultiple):
    template_name = 'richmin/widgets/select.html'

    def build_attrs(self, base_attrs, extra_attrs=None):
        extra_attrs['multiple'] = 'multiple'
        return {**base_attrs, **(extra_attrs or {})}

    @property
    def media(self):
        return forms.Media(
            css={'all': ('vendor/select2/css/select2.min.css',)},
            js=('vendor/select2/js/select2.min.js',),
        )



# Generated by Django 2.2.15 on 2020-10-14 12:37

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("books", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Library",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("address", models.CharField(max_length=255)),
                (
                    "librarian",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE, related_name="library", to=settings.AUTH_USER_MODEL
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="BookLoan",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text="Unique ID for this particular book across whole library",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("imprint", models.CharField(max_length=200)),
                ("due_back", models.DateField(blank=True, null=True)),
                ("loan_start", models.DateTimeField()),
                ("duration", models.DurationField(blank=True)),
                (
                    "status",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("m", "Maintenance"),
                            ("o", "On loan"),
                            ("a", "Available"),
                            ("r", "Reserved"),
                        ],
                        default="m",
                        help_text="Book availability",
                        max_length=1,
                    ),
                ),
                (
                    "book",
                    models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to="books.Book"),
                ),
                (
                    "borrower",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL
                    ),
                ),
            ],
            options={
                "ordering": ("due_back",),
            },
        ),
    ]



import pytest
from django.urls import reverse

from .test_app.library.factories import BookFactory, UserFactory
from .utils import parse_sidemenu


@pytest.mark.django_db
def test_no_delete_permission(client):
    """
    When our user has no delete permission, they dont see things they are not supposed to
    """
    user = UserFactory(permissions=['books.view_book'])
    book = BookFactory()

    url = reverse(
        'admin:books_book_change', args=(book.pk,)
    )
    delete_url = reverse(
        'admin:books_book_delete', args=(book.pk,)
    )
    client.force_login(user)

    response = client.get(url)
    assert delete_url not in response.content.decode()


@pytest.mark.django_db
def test_no_add_permission(client):
    """
    When our user has no add permission, they dont see things they are not supposed to
    """
    user = UserFactory(permissions=['books.view_book'])
    url = reverse('admin:books_book_changelist')
    add_url = reverse('admin:books_book_add')

    client.force_login(user)
    response = client.get(url)

    assert add_url not in response.content.decode()


@pytest.mark.django_db
def test_delete_but_no_view_permission(client):
    """
    When our user has delete but no view/change permission, menu items render out, but with no links

    As in Plain old Django Admin
    """
    user = UserFactory(permissions=['books.delete_book'])

    url = reverse('admin:index')
    client.force_login(user)

    response = client.get(url)
    assert parse_sidemenu(response) == {'Global': ['/en/admin/'], 'Books': [None]}


@pytest.mark.django_db
def test_no_permission(client):
    """
    When our user has no permissions at all, they see no menu or dashboard

    As in Plain old Django Admin
    """
    user = UserFactory(permissions=[])

    url = reverse('admin:index')
    client.force_login(user)

    response = client.get(url)
    assert parse_sidemenu(response) == {'Global': ['/en/admin/']}



from django.conf import settings
from django.conf.urls.i18n import i18n_patterns
from django.contrib import admin, messages
from django.contrib.auth import views as auth_views
from django.http import HttpResponseRedirect
from django.urls import include, path, re_path, reverse
from django.views.generic import RedirectView
from django.views.static import serve


def make_messages(request):
    messages.add_message(request, messages.INFO, 'Info message')
    messages.add_message(request, messages.ERROR, 'Error message')
    messages.add_message(request, messages.WARNING, 'Warning message')
    messages.add_message(request, messages.SUCCESS, 'Success message')

    return HttpResponseRedirect(reverse('admin:index'))


urlpatterns = [
    path('', RedirectView.as_view(pattern_name='admin:index', permanent=False)),
    path('admin/doc/', include('django.contrib.admindocs.urls')),
    path('make_messages/', make_messages, name='make_messages'),
    path('i18n/', include('django.conf.urls.i18n')),
]

urlpatterns += i18n_patterns(
    path('admin/password_reset/', auth_views.PasswordResetView.as_view(), name='admin_password_reset'),
    path('admin/password_reset/done/', auth_views.PasswordResetDoneView.as_view(), name='password_reset_done'),
    path('reset/<uidb64>/<token>/', auth_views.PasswordResetConfirmView.as_view(), name='password_reset_confirm'),
    path('reset/done/', auth_views.PasswordResetCompleteView.as_view(), name='password_reset_complete'),
    path('admin/', admin.site.urls),
)

if settings.DEBUG:
    urlpatterns.append(re_path(r'^static/(?P<path>.*)$', serve, kwargs={'document_root': settings.STATIC_ROOT}))

if 'debug_toolbar' in settings.INSTALLED_APPS:
    try:
        import debug_toolbar

        urlpatterns.append(path('__debug__/', include(debug_toolbar.urls)))
    except ImportError:
        pass



# Generated by Django 2.2.15 on 2020-10-14 12:37
from typing import List

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies: List[str] = []

    operations = [
        migrations.CreateModel(
            name="Author",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("first_name", models.CharField(max_length=100)),
                ("last_name", models.CharField(max_length=100)),
                ("date_of_birth", models.DateField(blank=True, null=True)),
                (
                    "date_of_death",
                    models.DateField(blank=True, null=True, verbose_name="Died"),
                ),
            ],
            options={
                "ordering": ("last_name", "first_name"),
            },
        ),
        migrations.CreateModel(
            name="Genre",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Enter a book genre (e.g. Science Fiction)",
                        max_length=200,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Book",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200)),
                (
                    "summary",
                    models.TextField(
                        help_text="Enter a brief description of the book",
                        max_length=1000,
                    ),
                ),
                (
                    "isbn",
                    models.CharField(
                        help_text='13 Character <a href="https://www.isbn-international.org/content/what-isbn">ISBN number</a>',
                        max_length=13,
                        verbose_name="ISBN",
                    ),
                ),
                ("published_on", models.DateField()),
                (
                    "author",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="books.Author",
                    ),
                ),
                (
                    "genre",
                    models.ManyToManyField(help_text="Select a genre for this book", to="books.Genre"),
                ),
            ],
        ),
    ]
